/*
 * Copyright (c) 2004-2007 Erez Zadok
 * Copyright (c) 2004-2007 Stony Brook University
 * Copyright (c) 2004-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 * Copyright (c) 2004-2005 Timothy Wong
 * Copyright (c) 2006-2007 Gopala Surya

 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */

#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include "thread_manager.h"
#include "res_manager.h"
#include "hash.h"
#include "hash_list.h"
#include "hashed_heap.h"


//Hash for PIDs to execution threads' rat location
static hashed_heap_t load_thread_loc_heap;
static hashed_heap_t free_thread_loc_heap;
static hash_list_t pid_to_thread_loc_hash;
static int cmd_counter = 0;

static int thread_data_compare(void* key1,int keylen1,
			       void* key2,int keylen2)
{
  int err = 0;
  int value1,value2;
  print_entry_location();
  assert(key1 && keylen1 > 0);
  assert(key2 && keylen2 > 0);

  value1 = *(int*)key1;
  value2 = *(int*)key2;

  if(value1 == value2)
    err = 0;
  else if(value1 > value2)
    err = -1;
  else
    err = 1;

 out:
  print_exit_location();
  return err;
}

void thread_manager_init()
{
  cmd_counter = 0;
  hash_list_init(&pid_to_thread_loc_hash,0);
  hashed_heap_init(&load_thread_loc_heap,0,thread_data_compare);
  hashed_heap_init(&free_thread_loc_heap,0,thread_data_compare);
}

void thread_manager_destroy()
{
  print_entry_location();

  hash_list_destroy(&pid_to_thread_loc_hash);
  hashed_heap_destroy(&load_thread_loc_heap);
  hashed_heap_destroy(&free_thread_loc_heap);

  print_exit_location();
  return;
}

static int _append_thread_to_rat(int pid)
{
  struct replay_command rcmd;
  int rat_loc;
  print_entry_location();

  rat_loc = res_manager_append_rat(pid,ARG_KTHREAD_DATA);

  print_exit_location();
  return rat_loc;
}

int thread_manager_request_thread(int pid)
{
  int loc = pid;
  thread_manager_data_t *t_data;

  print_entry_location();

  t_data = (thread_manager_data_t*)hash_list_find(&pid_to_thread_loc_hash,&pid,sizeof(pid));
  if(t_data) {
    t_data->last_use = cmd_counter;
    loc = t_data->pid;
    goto out;
  }

  t_data = malloc(sizeof(thread_manager_data_t));
  if(!t_data) {
    print_error_nomem();
    exit(ENOMEM);
  }
  memset(t_data,0,sizeof(thread_manager_data_t));
  t_data->pid = pid;
  t_data->first_use = t_data->last_use = cmd_counter;
  hash_list_add(&pid_to_thread_loc_hash,&pid,sizeof(pid),t_data);

  if(pid != 0) {
    //Add to heap, for sorting later
    hashed_heap_add(&load_thread_loc_heap,
		    &t_data->first_use,
		    sizeof(t_data->first_use),
		    t_data);
  }

 out:
  cmd_counter++;
  print_exit_location();
  return loc;
}

static void _heapify_last_use()
{
  thread_manager_data_t *t_data;
  int loc = 0;
  int size = 0;
  print_entry_location();

  size = hash_list_get_size(&pid_to_thread_loc_hash);

  for(loc = 0;loc < size;loc++) {
    t_data = hash_list_get_location(&pid_to_thread_loc_hash,loc);
    assert(t_data);
    if(t_data->pid != 0) {
      hashed_heap_add(&free_thread_loc_heap,
		      &t_data->last_use,
		      sizeof(t_data->last_use),
		      t_data);
    }
  }

  print_exit_location();
  return;
}

static hash_t* _resolve_reusable_threads()
{
  hash_t *pid_to_thread_hash = NULL;
  int create_value,free_value = -1;
  int thread_loc;
  int free_thread_loc = 0;
  thread_manager_data_t *load_entry = NULL;
  thread_manager_data_t *free_entry = NULL;

  print_entry_location();

  assert(hashed_heap_get_size(&load_thread_loc_heap) == hashed_heap_get_size(&free_thread_loc_heap));

  //Init the hash
  pid_to_thread_hash = malloc(sizeof(hash_t));
  if(!pid_to_thread_hash) {
    print_error_nomem();
    exit(ENOMEM);
  }
  hash_init(pid_to_thread_hash,0);

  while((load_entry = hashed_heap_remove_root(&load_thread_loc_heap))) {
    if(!free_entry) {
      free_entry = hashed_heap_remove_root(&free_thread_loc_heap);
      assert(free_entry);
      free_value = free_entry->last_use;
    }
    create_value = load_entry->first_use;

    if(load_entry->pid == free_entry->pid) {
      //Same object - try to reuse else add new
      thread_loc = free_thread_loc;
      free_thread_loc = 0;
      if(!thread_loc) {
	thread_loc = _append_thread_to_rat(load_entry->pid);
	assert(!free_thread_loc); //Shouldn't have anything in it
	free_thread_loc = thread_loc; //Re-add since it's the same obj
      }
      else
	load_entry->created = 1;
      load_entry->thread_loc = thread_loc;
      hash_add(pid_to_thread_hash,&load_entry->pid,sizeof(load_entry->pid),load_entry);
      free_entry = NULL;
      continue;
    }

    assert(free_value != -1);
    if(create_value < free_value) {
      //Try to reuse, if not, create thread
      thread_loc = free_thread_loc;
      free_thread_loc = 0;
      if(!thread_loc)
	thread_loc = _append_thread_to_rat(load_entry->pid);
      else
	load_entry->created = 1;
      load_entry->thread_loc = thread_loc;
      hash_add(pid_to_thread_hash,&load_entry->pid,sizeof(load_entry->pid),load_entry);
      continue;
    }

    //else - if(create value >= free_value)
    thread_loc = free_entry->thread_loc; //Set previously with load_entry->thread_loc
    assert(thread_loc != 0);
    load_entry->thread_loc = thread_loc;
    load_entry->created = 1;
    hash_add(pid_to_thread_hash,&load_entry->pid,sizeof(load_entry->pid),load_entry);
    free_entry = NULL;
    continue;
  }

 out:
  print_exit_location();
  return pid_to_thread_hash;
}

struct replay_command* _fill_create_thread_op(struct replay_command *srcmdp,int thread_loc,timeval_t time)
{
  print_entry_location();
  assert(srcmdp);

  memset(srcmdp,0,sizeof(struct replay_command));
  srcmdp->command = OP_CREATE_THREAD;
  srcmdp->time.secs = 0;
  srcmdp->time.usecs = 0;
  srcmdp->thread_loc = 0;
  srcmdp->ret = thread_loc;

  print_exit_location();
  return srcmdp;
}

static int _remap_prog_thread_loc(hash_t *pid_to_thread_hash)
{
  int err = 0;
  int thread_loc = 0;
  int pid = -1;
  int skip_increment = 0;
  struct prog_entry_s *prog_entry;
  struct prog_entry_s *prev_prog_entry = NULL;
  struct prog_entry_s *first_prog_entry = NULL;
  struct replay_command *srcmdp;
  thread_manager_data_t *t_data;
  print_entry_location();
  assert(pid_to_thread_hash);

  prog_entry = res_manager_get_prog_list();
  assert(prog_entry);

  while(prog_entry) {
    srcmdp = prog_entry->rcmdp;

    //Remap!
    //XXX: srcmdp->thread_loc is currently the pid, remap it to the correct thread_loc
    if(srcmdp->thread_loc == pid && pid != -1) {
      if(pid == 0)
	thread_loc = 0;
      else
	assert(thread_loc);
      srcmdp->thread_loc = thread_loc;
      goto next;
    }
    pid = srcmdp->thread_loc;
    if(pid == 0) {
      thread_loc = 0;
      srcmdp->thread_loc = thread_loc;
      goto next;
    }
    t_data = (thread_manager_data_t*)hash_find(pid_to_thread_hash,&pid,sizeof(pid));

    assert(t_data); //It should already have been hashed!
    assert(t_data->pid == pid);
    thread_loc = t_data->thread_loc;
    srcmdp->thread_loc = thread_loc;

    if(!t_data->created) {
      //not already created, insert!
      struct replay_command *new_srcmdp = malloc(sizeof(struct replay_command));
      if(!new_srcmdp) {
	print_error_nomem();
	exit(ENOMEM);
      }
      t_data->created = 1;

      if(prev_prog_entry == NULL) {
	//Replace head and add create op to the front
	first_prog_entry = prog_entry;
	memcpy(new_srcmdp,srcmdp,sizeof(struct replay_command));
	srcmdp = _fill_create_thread_op(srcmdp,t_data->thread_loc,srcmdp->time);
	prev_prog_entry = prog_entry; //Make prev the current
	prog_entry = prog_entry->next; //Skip the command we're about to add
	skip_increment = 1;
      }
      else {
	//Just fill the new replay command
	new_srcmdp = _fill_create_thread_op(new_srcmdp,t_data->thread_loc,srcmdp->time);
      }
      assert(new_srcmdp);
      res_manager_insert_prog(first_prog_entry,new_srcmdp);
    }

  next:
    if(!skip_increment) {
      prev_prog_entry = prog_entry;
      prog_entry = prog_entry->next;
    }
    else
      skip_increment = 0;
  }

  print_exit_location();
  return err;
}

int thread_manager_process_thread_requests()
{
  int err = 0;
  hash_t *pid_to_thread_hash;
  print_entry_location();

  _heapify_last_use();
  pid_to_thread_hash = _resolve_reusable_threads();
  if(!pid_to_thread_hash) {
    err = -1;
    goto out;
  }
  _remap_prog_thread_loc(pid_to_thread_hash);

 out:
  print_exit_location();
  return err;
}
